"""The learners module contains core classes and types for defining learner simulations.

This module contains the abstract interface expected for Learner implementations along
with a number of Learner implementations out of the box for testing and baseline comparisons.

TODO Add VowpalAdfLearner
"""

import math
import random

from abc import ABC, abstractmethod
from typing import Callable, Sequence, Tuple, Optional, Dict, cast, Generic, TypeVar
from itertools import accumulate
from collections import defaultdict

from coba.simulations import State, Action, Reward
from coba.utilities import check_vowpal_support
from coba.statistics import OnlineVariance

_S_in = TypeVar('_S_in', bound=State , contravariant=True)
_A_in = TypeVar('_A_in', bound=Action, contravariant=True)

class Learner(Generic[_S_in, _A_in], ABC):
    """The interface for Learner implementations."""

    @property
    @abstractmethod
    def name(self) -> str:
        """ The name of the learner.

        This value is used for descriptive purposes only when creating benchmark results.
        """
        ...

    @abstractmethod
    def choose(self, state: _S_in, actions: Sequence[_A_in]) -> int:
        """Choose which action to take.

        Args:
            state: The current state. This argument will be None when playing 
                a multi-armed bandit simulation and will contain context features 
                when playing a contextual bandit simulation. Context features could 
                be an individual number (e.g. 1.34), a string (e.g., "hot"), or a 
                tuple of strings and numbers (e.g., (1.34, "hot")) depending on the 
                simulation being played.
            actions: The current set of actions to choose from in the given state. 
                Action sets can be lists of numbers (e.g., [1,2,3,4]), a list of 
                strings (e.g. ["high", "medium", "low"]), or a list of lists such 
                as in the case of movie recommendations (e.g., [["action", "oscar"], 
                ["fantasy", "razzie"]]).

        Returns:
            An integer indicating the index of the selected action in the action set.
        """
        ...

    @abstractmethod
    def learn(self, state: _S_in, action: _A_in, reward: Reward) -> None:
        """Learn about the result of an action that was taken in a state.

        Args:
            state: The current state. This argument will be None when playing 
                a multi-armed bandit simulation and will contain context features 
                when playing a contextual bandit simulation. Context features could 
                be an individual number (e.g. 1.34), a string (e.g., "hot"), or a 
                tuple of strings and numbers (e.g., (1.34, "hot")) depending on the 
                simulation being played.
            action: The action that was selected to play and observe its reward. 
                An Action can be an individual number (e.g., 2), a string (e.g. 
                "medium"), or a list of some combination of numbers or strings
                (e.g., ["action", "oscar"]).
            reward: the reward received for taking the given action in the given state.
        """
        ...

class LambdaLearner(Learner[_S_in, _A_in]):
    """A Learner implementation that chooses and learns according to provided lambda functions."""

    def __init__(self, 
                 choose: Callable[[_S_in, Sequence[_A_in]], int], 
                 learn : Optional[Callable[[_S_in, _A_in, Reward],None]] = None,
                 name  : str = "Lambda") -> None:
        """Instantiate LambdaLearner.

        Args:
            chooser: a function matching the super().choose() signature. All parameters are passed straight through.
            learner: a function matching the super().learn() signature. If provided all parameters are passed
                straight through. If the function isn't provided then no learning occurs.
        """
        self._choose = choose
        self._learn  = learn
        self._name   = name

    @property
    def name(self) -> str:
        """The name of the Learner.
        
        See the base class for more information
        """  
        return self._name

    def choose(self, state: _S_in, actions: Sequence[_A_in]) -> int:
        """Choose via the provided lambda function.

        Args:
            state: The state we're currently in. See the base class for more information.
            actions: The actions to choose from. See the base class for more information.

        Returns:
            The index of the selected action. See the base class for more information.
        """

        return self._choose(state, actions)

    def learn(self, state: _S_in, action: _A_in, reward: Reward) -> None:
        """Learn via the optional lambda function or learn nothing without a lambda function.

        Args:
            state: The state we're learning about. See the base class for more information.
            action: The action that was selected in the state. See the base class for more information.
            reward: The reward that was gained from the action. See the base class for more information.
        """
        if self._learn is None:
            pass
        else:
            self._learn(state,action,reward)

class RandomLearner(Learner[State, Action]):
    """A Learner implementation that selects an action at random and learns nothing."""

    @property
    def name(self) -> str:
        """The name of the Learner.

        See the base class for more information
        """  
        return "Random"

    def choose(self, state: State, actions: Sequence[Action]) -> int:
        """Choose a random action from the action set.
        
        Args:
            state: The state we're currently in. See the base class for more information.
            actions: The actions to choose from. See the base class for more information.

        Returns:
            The index of the selected action. See the base class for more information.
        """
        return random.randint(0, len(actions)-1)

    def learn(self, state: State, action: Action, reward: Reward) -> None:
        """Learns nothing.

        Args:
            state: The state we're learning about. See the base class for more information.
            action: The action that was selected in the state. See the base class for more information.
            reward: The reward that was gained from the action. See the base class for more information.
        """

        pass

class EpsilonLearner(Learner[State, Action]):
    """A learner using epsilon-greedy searching while smoothing observations into a state/state-action lookup table.

    Remarks:
        This algorithm does not use any function approximation to attempt to generalize observed rewards.
    """

    def __init__(self, epsilon: float, default: Optional[float] = None, include_state: bool = False) -> None:
        """Instantiate an EpsilonLearner.

        Args:
            epsilon: A value between 0 and 1. We explore with probability epsilon and exploit otherwise.
            default: Our initial guess of the expected rewards for all state-action pairs.
            include_state: If true lookups are a function of state-action otherwise they are a function of action.
        """

        self._epsilon       = epsilon
        self._include_state = include_state
        self._N: Dict[Tuple[State, Action], int            ] = defaultdict(lambda: int(0 if default is None else 1))
        self._Q: Dict[Tuple[State, Action], Optional[float]] = defaultdict(lambda: default)

    @property
    def name(self) -> str:
        """The name of the Learner.

        See the base class for more information
        """
        return "Epislon-greedy"

    def choose(self, state: State, actions: Sequence[Action]) -> int:
        """Choose greedily with probability 1-epsilon. Choose a randomly with probability epsilon.
        
        Args:
            state: The state we're currently in. See the base class for more information.
            actions: The actions to choose from. See the base class for more information.

        Returns:
            The index of the selected action. See the base class for more information.
        """
        if(random.random() <= self._epsilon): return random.randint(0,len(actions)-1)

        keys        = [ self._key(state,action) for action in actions ]
        values      = [ self._Q[key] for key in keys ]
        max_value   = None if set(values) == {None} else max(v for v in values if v is not None)
        max_indexes = [i for i in range(len(values)) if values[i]==max_value]

        return random.choice(max_indexes)

    def learn(self, state: State, action: Action, reward: Reward) -> None:
        """Smooth the observed reward into our current estimate of either E[R|S,A] or E[R|A].

        Args:
            state: The state we're learning about. See the base class for more information.
            action: The action that was selected in the state. See the base class for more information.
            reward: The reward that was gained from the action. See the base class for more information.
        """

        key   = self._key(state,action)
        alpha = 1/(self._N[key]+1)

        old_Q = cast(float, 0 if self._Q[key] is None else self._Q[key])

        self._Q[key] = (1-alpha) * old_Q + alpha * reward
        self._N[key] = self._N[key] + 1

    def _key(self, state: State, action: Action) -> Tuple[State,Action]:
        return (state, action) if self._include_state else (None, action)

class VowpalLearner(Learner[State, Action]):
    """A learner using Vowpal Wabbit's contextual bandit command line interface.

    Remarks:
        This learner requires that the Vowpal Wabbit package be installed. This package can be
        installed via `pip install vowpalwabbit`. To learn more about solving contextual bandit
        problems with Vowpal Wabbit see https://vowpalwabbit.org/tutorials/contextual_bandits.html.
    """

    def __init__(self, epsilon: Optional[float] = 0.1, bag: Optional[int] = None, cover: Optional[int] = None) -> None:
        """Instantiate a VowpalLearner.

        Args:
            epsilon: A value between 0 and 1. If provided exploration will follow epsilon-greedy.
            bag: An integer value greater than 0. This value determines how many separate policies will be
                learned. Each policy will be learned from bootstrap aggregation making each policy unique. 
                For each choice one policy will be selected according to a uniform distribution and followed.
            cover: An integer value greater than 0. This value value determines how many separate policies will be
                learned. These policies are learned in such a way to explicitly optimize policy diversity in order
                to control exploration. For each choice one policy will be selected according to a uniform distribution
                and followed. For more information on this algorithm see Agarwal et al. (2014).

        Remarks:
            Only one parameter of epsilon, bag and cover should be set. If more than one parameter is set then 
            only one value is used according to the precedence of first use cover then bag then epsilon.

        References:
            Agarwal, Alekh, Daniel Hsu, Satyen Kale, John Langford, Lihong Li, and Robert Schapire. "Taming 
            the monster: A fast and simple algorithm for contextual bandits." In International Conference on 
            Machine Learning, pp. 1638-1646. 2014.
        """

        check_vowpal_support('VowpalLearner.__init__')
        from vowpalwabbit import pyvw #type: ignore #ignored due to mypy error
        self._vw = pyvw.vw

        if epsilon is not None:
            self._explore = f"--epsilon {epsilon}"

        if bag is not None:
            self._explore = f"--bag {bag}"

        if cover is not None:
            self._explore = f"--cover {cover}"

        self._actions: Sequence[State]                  = []
        self._prob   : Dict[Tuple[State,Action], float] = {}

    @property
    def name(self) -> str:
        """The name of the Learner.
        
        See the base class for more information
        """  
        return "Vowpal"

    def choose(self, state: State, actions: Sequence[Action]) -> int:
        """Choose an action according to the explor-exploit parameters passed into the contructor.

        Args:
            state: The state we're currently in. See the base class for more information.
            actions: The actions to choose from. See the base class for more information.

        Returns:
            The index of the selected action. See the base class for more information.

        Remarks:
            We assume that the action set passed in is always the same. This restriction
            is forced on us by Vowpal Wabbit. If your action set is not static then you
            should use VowpalAdfLearner.
        """

        if len(self._actions) == 0:
            self._actions = actions
            self._vw_learner = self._vw(f"--cb_explore {len(actions)} -q UA  {self._explore} --quiet")

        pmf = self._vw_learner.predict("| " + self._vw_format(state))

        cdf   = list(accumulate(pmf))
        rng   = random.random()
        index = [ rng < c for c in cdf].index(True)

        self._prob[(state, actions[index])] = pmf[index]

        return index

    def learn(self, state: State, action: Action, reward: Reward) -> None:
        """Learn from the obsered reward for the given state action pair.

        Args:
            state: The state we're learning about. See the base class for more information.
            action: The action that was selected in the state. See the base class for more information.
            reward: The reward that was gained from the action. See the base class for more information.
        """
        
        prob  = self._prob[(state,action)]
        cost  = -reward

        vw_state  = self._vw_format(state)
        vw_action = str(self._actions.index(action)+1)

        self._vw_learner.learn( vw_action + ":" + str(cost) + ":" + str(prob) + " | " + vw_state)

    def _vw_format(self, state: State) -> str:
        """convert state into the proper format for pyvw.
        
        Args:
            state: The state we wish to convert to pyvw representation.

        Returns:
            The state in pyvw representation.
        """
        if state is None:  return ""

        if isinstance(state, (int,float,str)):
            return str(state)

        #Right now if a state isn't one of the above types it
        #has to be a tuple. The type checker doesn't know that,
        #however, so we tell it with the explicit cast below.
        #During runtime this cast will do nothing.
        return " ". join(map(str, cast(tuple,state)))

class UcbTunedLearner(Learner[State, Action]):
    """This is an implementation of Auer et al. (2002) UCB1-Tuned algorithm.

    References:
        Auer, Peter, Nicolo Cesa-Bianchi, and Paul Fischer. "Finite-time analysis of 
        the multiarmed bandit problem." Machine learning 47.2-3 (2002): 235-256.
    """
    def __init__(self):
        """Instantiate a UcbTunedLearner."""

        self._init_a: int = 0
        self._t     : int = 0
        self._s     : Dict[Action,int] = {}
        self._m     : Dict[Action,float] = {}
        self._v     : Dict[Action,OnlineVariance] = defaultdict(OnlineVariance)

    @property
    def name(self) -> str:
        """The name of the Learner.

        See the base class for more information
        """
        return "UCB"

    def choose(self, state: State, actions: Sequence[Action]) -> int:
        """Choose an action greedily according to the upper confidence bound estimates.

        Args:
            state: The state we're currently in. See the base class for more information.
            actions: The actions to choose from. See the base class for more information.

        Returns:
            The index of the selected action. See the base class for more information.
        """
        #we initialize by playing every action once
        if self._init_a < len(actions):
            self._init_a += 1
            return self._init_a-1

        else:
            values      = [ self._m[a] + self._Avg_R_UCB(a) if a in self._m else None for a in actions ]
            max_value   = None if set(values) == {None} else max(v for v in values if v is not None)
            max_indexes = [i for i in range(len(values)) if values[i]==max_value]
            return random.choice(max_indexes)

    def learn(self, state: State, action: Action, reward: Reward) -> None:
        """Smooth the observed reward into our current estimate of E[R|S,A].

        Args:
            state: The state we're learning about. See the base class for more information.
            action: The action that was selected in the state. See the base class for more information.
            reward: The reward that was gained from the action. See the base class for more information.
        """
        if action not in self._s:
            self._s[action] = 1
        else:
            self._s[action] += 1

        if action not in self._m:
            self._m[action] = reward
        else:
            self._m[action] = (1-1/self._s[action]) * self._m[action] + 1/self._s[action] * reward

        self._t         += 1
        self._s[action] += 1
        self._v[action].update(reward)

    def _Avg_R_UCB(self, action: Action) -> float:
        """Produce the estimated upper confidence bound (UCB) for E[R|A].

        Args:
            action: The action for which we want to retrieve UCB for E[R|A].

        Returns:
            The estimated UCB for E[R|A].

        Remarks:
            See the beginning of section 4 in the algorithm's paper for this equation.
        """
        ln = math.log; n = self._t; n_j = self._s[action]; V_j = self._Var_R_UCB(action)

        return math.sqrt(ln(n)/n_j * min(1/4,V_j))

    def _Var_R_UCB(self, action: Action) -> float:
        """Produce the upper confidence bound (UCB) for Var[R|A].

        Args:
            action: The action for which we want to retrieve UCB for Var[R|A].

        Returns:
            The estimated UCB for Var[R|A].

        Remarks:
            See the beginning of section 4 in the algorithm's paper for this equation.
        """
        ln = math.log; t = self._t; s = self._s[action]; var = self._v[action].variance

        return var + math.sqrt(2*ln(t)/s)


"""The contexts module contains classes determining the context for shared functionality.

TODO Add unittests for CobaConfig.
TODO Add unittests for all CacheInterface implementations.
"""

import json
import copy
import collections

from itertools import repeat
from gzip import compress, decompress
from abc import ABC, abstractmethod
from io import BytesIO
from typing import Union, Generic, TypeVar, Dict, IO, Mapping, Any, Optional, List, MutableMapping, cast
from pathlib import Path

_K = TypeVar("_K")
_V = TypeVar("_V")    

class TemplatingEngine():

    @staticmethod
    def parse(json_val:Union[str, Any]):

        root = json.loads(json_val) if isinstance(json_val, str) else json_val

        if "templates" in root:

            templates: Dict[str,Dict[str,Any]] = root.pop("templates")
            nodes    : List[Any]               = [root]
            scopes   : List[Dict[str,Any]]     = [{}]

            def materialize_template(document: MutableMapping[str,Any], template: Mapping[str,Any]):

                for key in template:
                    if key in document:
                        if isinstance(template[key], collections.Mapping) and isinstance(template[key], collections.Mapping):
                            materialize_template(document[key],template[key])
                    else:
                        document[key] = template[key]

            def materialize_variables(document: MutableMapping[str,Any], variables: Mapping[str,Any]):
                for key in document:
                    if isinstance(document[key],str) and document[key] in variables:
                        document[key] = variables[document[key]]

            while len(nodes) > 0:
                node  = nodes.pop()
                scope = scopes.pop().copy()  #this could absolutely be made more memory-efficient if needed

                if isinstance(node, collections.MutableMapping):

                    if "template" in node and node["template"] not in templates:
                        raise Exception(f"We were unable to find template '{node['template']}'.")

                    keys      = list(node.keys())
                    template  = templates[node.pop("template")] if "template" in node else cast(Dict[str,Any], {})
                    variables = { key:node.pop(key) for key in keys if key.startswith("$") }

                    template = copy.deepcopy(template)
                    scope.update(variables)

                    materialize_template(node, template)
                    materialize_variables(node, scope)

                    for child_node, child_scope in zip(node.values(), repeat(scope)):
                        nodes.append(child_node)
                        scopes.append(child_scope)

                if isinstance(node, collections.Sequence) and not isinstance(node, str):
                    for child_node, child_scope in zip(node, repeat(scope)):
                        nodes.append(child_node)
                        scopes.append(child_scope)

        return root

class CobaConfig():
    def __init__(self):
        search_paths = [Path("./.coba"), Path.home() / ".coba"]

        config = {}

        for potential_path in search_paths:
            if potential_path.exists():
                with open(potential_path) as fs:
                    config = json.load(fs)
                break

        self._config = config

    @property
    def openml_api_key(self) -> Optional[str]:
        return self._config.get("openml_api_key", None)

    @property
    def file_cache(self) -> Mapping[str,Any]:
        return self._config.get("file_cache", {"type":"none"})        

class CacheInterface(Generic[_K, _V], ABC):

    @abstractmethod
    def __contains__(self, key: _K) -> bool:
        ...

    @abstractmethod
    def get(self, key: _K) -> _V:
        ...

    @abstractmethod
    def put(self, key: _K, value: _V) -> _V:
        ...

    @abstractmethod
    def rmv(self, key: _K) -> None:
        ...

class NoneCache(CacheInterface[_K, _V]):
    def __init__(self) -> None:
        self._cache: Dict[_K,_V] = {}

    def __contains__(self, key: _K) -> bool:
        return False

    def get(self, key: _K) -> _V:
        raise Exception("the key didn't exist in the cache")

    def put(self, key: _K, value: _V) -> _V:
        return value

    def rmv(self, key: _K):
        pass

class MemoryCache(CacheInterface[_K, _V]):
    def __init__(self) -> None:
        self._cache: Dict[_K,_V] = {}

    def __contains__(self, key: _K) -> bool:
        return key in self._cache

    def get(self, key: _K) -> _V:
        return self._cache[key]

    def put(self, key: _K, value: _V) -> _V:
        self._cache[key] = value

        return value

    def rmv(self, key: _K):
        del self._cache[key]

class DiskCache(CacheInterface[str, IO[bytes]]):
    def __init__(self, path: Union[str, Path]) -> None:
        self._cache_dir = path if isinstance(path, Path) else Path(path).expanduser()

    def __contains__(self, filename: str) -> bool:
        is_gzip = filename.endswith(".gz")

        gzip_filename = filename + ("" if is_gzip else ".gz")

        return (self._cache_dir/gzip_filename).exists()

    def get(self, filename: str) -> IO[bytes]:
        is_gzip = filename.endswith(".gz")

        gzip_filename = filename + ("" if is_gzip else ".gz")
        gzip_bytes    = (self._cache_dir/gzip_filename).read_bytes()

        return BytesIO(gzip_bytes if is_gzip else decompress(gzip_bytes))

    def put(self, filename: str, value: IO[bytes]) -> IO[bytes]:
        is_gzip = filename.endswith(".gz")

        gzip_filename = filename + ("" if is_gzip else ".gz")
        gzip_bytes    = value.read() if is_gzip else compress(value.read())

        (self._cache_dir/gzip_filename).parent.mkdir(parents=True, exist_ok=True)
        (self._cache_dir/gzip_filename).touch()
        (self._cache_dir/gzip_filename).write_bytes(gzip_bytes)

        return self.get(filename)

    def rmv(self, filename: str) -> None:
        is_gzip = filename.endswith(".gz")

        gzip_filename = filename + ("" if is_gzip else ".gz")

        (self._cache_dir/gzip_filename).unlink()

class ExecutionContext:
    """Create a global execution context to allow easy mocking and modification.

    In short, So long as the same modulename is always used to import and the import
    always occurs on the same thread I'm fairly confident this pattern will always work.

    In long, I'm somewhat unsure about this pattern for the following reasons:
        > While there seems concensus that multi-import doesn't repeat [1] it may if different modulenames are used [2]
            [1] https://stackoverflow.com/a/19077396/1066291
            [2] https://stackoverflow.com/q/13392038/1066291

        > Python 3.7 added in explicit context management in [3,4] but this implementation is thread local only
            [3] https://www.python.org/dev/peps/pep-0567/
            [4] https://docs.python.org/3/library/contextvars.html
    """

    TemplatingEngine: TemplatingEngine = TemplatingEngine()
    CobaConfig      : CobaConfig     = CobaConfig()
    FileCache       : CacheInterface = NoneCache()

    if CobaConfig.file_cache["type"] == "disk":
        FileCache = DiskCache(CobaConfig.file_cache["directory"])

    if CobaConfig.file_cache["type"] == "memory":
        FileCache  = MemoryCache()


"""The statistics module contains algorithms and methods to calculate statistics.

TODO Add unittests to make sure SummaryStats.blend works correctly
TODO Add unittests to make sure SummaryStats.blend_cumulative works correctly
"""

import math

from typing import Sequence, Iterable, List

class OnlineVariance():
    """Calculate sample variance in an online fashion.
    
    Remarks:
        This algorithm is known as Welford's algorithm and the implementation below
        is a modified version of the Python algorithm by Wikepedia contirubtors (2020).

    References:
        Wikipedia contributors. (2020, July 6). Algorithms for calculating variance. In Wikipedia, The
        Free Encyclopedia. Retrieved 18:00, July 24, 2020, from 
        https://en.wikipedia.org/w/index.php?title=Algorithms_for_calculating_variance&oldid=966329915
    """

    def __init__(self) -> None:
        """Instatiate an OnlineVariance calcualator."""
        self._count    = 0.
        self._mean     = 0.
        self._M2       = 0.
        self._variance = float("nan")

    @property
    def variance(self) -> float:
        """The variance of all given updates."""
        return self._variance

    def update(self, value: float) -> None:
        """Update the current variance with the given value."""
        
        (count,mean,M2) = (self._count, self._mean, self._M2)

        count   += 1
        delta   = value - mean
        mean   += delta / count
        delta2  = value - mean
        M2     += delta * delta2

        (self._count, self._mean, self._M2) = (count, mean, M2)

        if count > 1:
            self._variance = M2 / (count - 1)

class OnlineMean():
    """Calculate mean in an online fashion."""

    def __init__(self):
        self._n = 0
        self._mean = float('nan')

    @property
    def mean(self) -> float:
        """The mean of all given updates."""

        return self._mean

    def update(self, value:float) -> None:
        """Update the current mean with the given value."""
        
        self._n += 1

        alpha = 1/self._n

        self._mean = value if alpha == 1 else (1 - alpha) * self._mean + alpha * value

class SummaryStats:
    """A class to store summary statistics calculated from some sample."""

    @staticmethod
    def from_observations(observations: Sequence[float]) -> 'SummaryStats':
        return SummaryStats().add_observations(observations)

    @staticmethod
    def blend(stats: Iterable['SummaryStats']) -> 'SummaryStats':
        """Calculate the resulting stats from blending the SummaryStats of several samples.

        Args:
            stats: The previously calculated stats that we wish to merge.

        Remarks:
            In theory, if every 'Stats' object kept their entire sample then we could estimate all the
            values that we calculate below empirically by mixing all samples into a single big pot and
            then recalculating mean, variance and SEM. Unfortunately, this approach while simple has two
            drawbacks. First, if a benchmark has many simulations with many interactions this means that 
            our package will always be constrained since our memory would be O(n) to interactions. Second, our 
            computation complexity would also be O(n) to interactions since merging stats would require relooping 
            over all samples again. One downside of not taking the simple approach is that stats with N == 1 
            become problematic since we don't know their variance. Therefore we adjust for this below by 
            adding the average variance back in for ever stat with N == 1.
        """

        total_sum = 0.
        total_var = 0.
        total_N   = 0
        total_N_1 = 0 

        for stat in stats: #type: ignore #(this error is a bug with pylance)
            total_sum += stat.mean * stat.N if stat.N > 0 else 0
            total_var += (stat.SEM**2)*(stat.N**2) if stat.N > 1 else 0
            total_N   += stat.N
            total_N_1 += int(stat.N == 1)

        #when we only have a single observation there is no way for us to estimate
        #the variance of that random number. Therefore we add a neutral amount of
        #variance since this number is still in total and total_N and we don't want
        #to remove it.
        total_var += (total_var/total_N) * total_N_1

        N        = total_N
        mean     = total_sum/total_N
        variance = float('nan') # once we start to blend samples there's no longer a meaningful way to calculate variance
        SEM      = math.sqrt(total_var/total_N**2)

        return SummaryStats(N, mean, variance, SEM)

    @staticmethod
    def blend_cumulative(stats: Iterable['SummaryStats']) -> Sequence['SummaryStats']:
        """Calculate the resulting stats from cumulatively blending the SummaryStats of several samples.

        Args:
            stats: The previously calculated stats that we wish to blend cumulatively.

        Remarks:
            For more information see `blend`.
        """

        total_sum = 0.
        total_var = 0.
        total_N   = 0

        cumulative_stats: List[SummaryStats] = []

        for stat in stats: #type: ignore #(this error is a bug with pylance)
            total_N   += stat.N
            total_sum += stat.mean * stat.N if stat.N > 0 else 0
            total_var += (stat.SEM**2)*(stat.N**2) if stat.N > 1 else total_var/total_N

            N    = total_N
            mean = total_sum/total_N
            SEM  = math.sqrt(total_var/total_N**2)

            cumulative_stats.append(SummaryStats(N=N, mean=mean, SEM=SEM))

        return cumulative_stats



    def __init__(self, N: int = 0, mean: float = float('nan'), variance:float = float('nan'), SEM: float = float('nan')):
        """Instantiate a Stats class.

        Args:
            N: The size of the sample of interest.
            mean: The mean for some sample of interest.
            variance: The variance for some sample of interest.
            SEM: The standard error of the mean for some sample of interest.
        """

        self._online_mean     = OnlineMean()
        self._online_variance = OnlineVariance()

        self._N    = N
        self._mean = mean
        self._var  = variance
        self._SEM  = SEM

    @property
    def N(self) -> int:
        """The size of the sample."""
        return self._N

    @property
    def mean(self) -> float:
        """The mean for some sample."""
        return self._mean

    @property
    def variance(self) -> float:
        """The mean for some sample."""
        return self._var

    @property
    def SEM(self) -> float:
        """The mean for some sample."""
        return self._SEM

    def add_observations(self, observations: Sequence[float]) -> 'SummaryStats':

        for observation in observations:
            self._online_mean.update(observation)
            self._online_variance.update(observation)

        self._N    += len(observations)
        self._mean  = self._online_mean.mean
        self._var   = self._online_variance.variance
        self._SEM   = math.sqrt(self._var/self._N) if self._N > 0 else float('nan')

        return self


"""The benchmarks module contains core benchmark functionality and protocols.

This module contains the abstract interface expected for Benchmark implementations. This 
module also contains several Benchmark implementations and Result data transfer class.

TODO Finish adding docstrings to Result
"""

import json
import collections

from abc import ABC, abstractmethod
from typing import Union, Sequence, List, Callable, Generic, TypeVar, Dict, Any, overload, cast
from itertools import islice, count
from statistics import median

from coba.simulations import Interaction, LazySimulation, Simulation, State, Action
from coba.learners import Learner
from coba.contexts import ExecutionContext
from coba.statistics import SummaryStats

_S = TypeVar('_S', bound=State)
_A = TypeVar('_A', bound=Action)

class Result:

    def __init__(self,
        learner_name:str,
        simulation_index:int,
        batch_index: int,
        interaction_count: int,
        median_feature_count: int,
        median_action_count: int,
        stats: SummaryStats) -> None:

        self._learner_name         = learner_name
        self._simulation_index     = simulation_index
        self._batch_index          = batch_index
        self._median_feature_count = median_feature_count
        self._median_action_count  = median_action_count
        self._interaction_count    = interaction_count
        self._stats                = stats

    @property
    def learner_name(self) -> str:
        return self._learner_name

    @property
    def simulation_index(self) -> int:
        return self._simulation_index

    @property
    def batch_index(self) -> int:
        return self._batch_index

    @property
    def interaction_count(self) -> int:
        return self._interaction_count

    @property
    def median_feature_count(self) -> int:
        return self._median_feature_count

    @property
    def median_action_count(self) -> int:
        return self._median_action_count

    @property
    def stats(self) -> SummaryStats:
        return self._stats

    def to_dict(self) -> Dict[str, Any]:
        return {
            "learner_name"        : self._learner_name,
            "simulation_index"    : self._simulation_index,
            "batch_index"         : self._batch_index,
            "interaction_count"   : self._interaction_count,
            "median_feature_count": self._median_feature_count,
            "median_action_count" : self._median_action_count,
            "stats"               : self._stats
        }

class Benchmark(Generic[_S,_A], ABC):
    """The interface for Benchmark implementations."""

    @abstractmethod
    def evaluate(self, learner_factories: Sequence[Callable[[],Learner[_S,_A]]]) -> Sequence[Result]:
        """Calculate the performance for a provided bandit Learner.

        Args:
            learner_factories: A sequence of functions to create Learner instances. Each function 
                should always create the same Learner in order to get an unbiased performance 
                Result. This method can be as simple as `lambda: MyLearner(...)`.

        Returns:
            The resulting performance statistics for each given learner to evaluate.

        Remarks:
            The learner factory is necessary because a Result can be calculated using
            observed performance over several simulations. In these cases the easiest 
            way to reset a learner's learned policy is to create a new learner.
        """
        ...

class UniversalBenchmark(Benchmark[_S,_A]):
    """An on-policy Benchmark using samples drawn from simulations to estimate performance statistics."""

    @staticmethod
    def from_json(json_val:Union[str, Dict[str,Any]]) -> 'UniversalBenchmark':
        """Create a UniversalBenchmark from configuration IO.

        Args:
            json_val: Either a json string or the decoded json object.

        Returns:
            The UniversalBenchmark representation of the given JSON string or object.
        """

        if isinstance(json_val, str):
            config = cast(Dict[str,Any],json.loads(json_val))
        else:
            config = json_val

        config = ExecutionContext.TemplatingEngine.parse(config)

        is_singular = isinstance(config["simulations"], dict)
        sim_configs = config["simulations"] if not is_singular else [ config["simulations"] ]

        #by default load simulations lazily
        for sim_config in sim_configs:
            if "lazy" not in sim_config:
                sim_config["lazy"] = True

        simulations = [ Simulation.from_json(sim_config) for sim_config in sim_configs ]

        if "count" in config["batches"]:
            return UniversalBenchmark(simulations, batch_count=config["batches"]["count"])
        else:
            return UniversalBenchmark(simulations, batch_size=config["batches"]["size"])    

    @overload
    def __init__(self, 
        simulations: Sequence[Simulation[_S,_A]],
        *, 
        batch_count: int) -> None:
        ...

    @overload
    def __init__(self, 
        simulations: Sequence[Simulation[_S,_A]],
        *, 
        batch_size: Union[int, Sequence[int], Callable[[int],int]]) -> None:
        ...

    def __init__(self,
        simulations: Sequence[Simulation[_S,_A]], 
        batch_count: int = None, 
        batch_size : Union[int, Sequence[int], Callable[[int],int]] = None) -> None:
        """Instantiate a UniversalBenchmark.

        Args:
            simulations: A sequence of simulations to benchmark against.
            batch_count: How many interaction batches per simulation (batch_size will be spread evenly).
            batch_size: An indication of how large every batch should be. If batch_size is an integer
                then simulations will run until completion with batch sizes of the given int. If 
                batch_size is a sequence of integers then `sum(batch_size)` interactions will be 
                pulled from simulations and batched according to the sequence. If batch_size is a 
                function then simulation run until completion with batch_size determined by function.
        """

        self._simulations = simulations
        self._batch_count = batch_count
        self._batch_size  = batch_size

    def evaluate(self, learner_factories: Sequence[Callable[[],Learner[_S,_A]]]) -> Sequence[Result]:
        """Collect observations of a Learner playing the benchmark's simulations to calculate Results.

        Args:
            learner_factories: See the base class for more information.

        Returns:
            See the base class for more information.
        """

        def feature_count(i: Interaction) -> int:
            return len(i.state) if isinstance(i.state,tuple) else 0 if i.state is None else 1

        def action_count(i: Interaction) -> int:
            return len(i.actions)

        results: List[Result] = []

        for simulation_index, simulation in enumerate(self._simulations):
            try:

                if isinstance(simulation, LazySimulation):
                    simulation.load()

                interaction_count    = len(simulation.interactions)
                median_feature_count = cast(int,median([feature_count(i) for i in simulation.interactions]))
                median_action_count  = cast(int,median([action_count(i) for i in simulation.interactions]))

                batch_sizes    = self._batch_sizes(len(simulation.interactions))
                n_interactions = sum(batch_sizes)

                for learner_index, factory in enumerate(learner_factories):

                    learner       = factory()
                    batch_index   = 0
                    batch_choices = []

                    for r in islice(simulation.interactions, n_interactions):

                        action_index = learner.choose(r.state, r.actions)
                        batch_choices.append(r.choices[action_index])

                        if len(batch_choices) == batch_sizes[batch_index]:
                            stats = SummaryStats()
                            obs   = simulation.rewards(batch_choices)
                            name  = self._safe_name(learner_index, learner)

                            for (state,action,reward) in obs:
                                learner.learn(state,action,reward)
                                stats.add_observations([reward])

                            results.append(Result(
                                name, 
                                simulation_index, 
                                batch_index, 
                                interaction_count, 
                                median_feature_count, 
                                median_action_count,
                                stats
                            ))

                            batch_choices = []
                            batch_index += 1

            except Exception as e:
                print(f"     * {e}")

            if isinstance(simulation, LazySimulation):
                simulation.unload()

        return results

    def _batch_sizes(self, n_interactions: int) -> Sequence[int]:

        if self._batch_count is not None:
            
            batches   = [int(float(n_interactions)/(self._batch_count))] * self._batch_count
            remainder = n_interactions % self._batch_count
            
            if remainder > 0:
                spacing = float(self._batch_count)/remainder
                for i in range(remainder): batches[int(i*spacing)] += 1

            return batches
        
        if isinstance(self._batch_size, int): 
            return [self._batch_size] * int(float(n_interactions)/self._batch_size)

        if isinstance(self._batch_size, collections.Sequence): 
            return self._batch_size

        if callable(self._batch_size):
            batch_size_iter        = (self._batch_size(i) for i in count())
            next_batch_size        = next(batch_size_iter)
            remaining_interactions = n_interactions
            batch_sizes: List[int] = []

            while remaining_interactions > next_batch_size:
                batch_sizes.append(next_batch_size)
                remaining_interactions -= next_batch_size
                next_batch_size  = next(batch_size_iter)
            
            return batch_sizes
        
        raise Exception("We were unable to determine batch size from the supplied parameters")

    def _safe_name(self, learner_index:int, learner: Learner) -> str:
        try:
            return learner.name
        except:
            return str(learner_index)


"""The analysis module contains functionality to help with analyzing Results."""

from collections import defaultdict
from itertools import groupby
from typing import Sequence, Callable, Tuple, List, Dict

from coba.statistics import SummaryStats
from coba.utilities import check_matplotlib_support
from coba.benchmarks import Result

class Plots():

    @staticmethod
    def standard_plot(results: Sequence[Result]) -> None:

        def plot_stats(axes, label, stats):
            x = [ i+1           for i in range(len(stats)) ]
            y = [ i.mean        for i in stats             ]
            l = [ i.mean-i.SEM  for i in stats             ]
            u = [ i.mean+i.SEM  for i in stats             ]

            axes.plot(x,y, label=label)
            axes.fill_between(x, l, u, alpha = 0.25)

        learner_stats: Dict[str,List[SummaryStats]] = defaultdict(list)

        group_key: Callable[[Result], Tuple[str,int]] = lambda r: (r.learner_name, r.batch_index)
        for batch_group in groupby(sorted(results, key=group_key), key=group_key):
            learner_name = batch_group[0][0]
            batch_stats  = [result.stats for result in  batch_group[1]]
            learner_stats[learner_name].append(SummaryStats.blend(batch_stats))

        check_matplotlib_support('Plots.standard_plot')
        import matplotlib.pyplot as plt #type: ignore

        fig = plt.figure()

        ax1 = fig.add_subplot(1,2,1) #type: ignore
        ax2 = fig.add_subplot(1,2,2) #type: ignore

        for learner_name, stats in learner_stats.items(): plot_stats(ax1, learner_name, stats)

        ax1.set_title("Reward by Batch Index")
        ax1.set_ylabel("Mean Reward")
        ax1.set_xlabel("Batch Index")

        for learner_name, stats in learner_stats.items(): plot_stats(ax2, learner_name, SummaryStats.blend_cumulative(stats))

        ax2.set_title("Progressive Validation Reward")
        ax2.set_xlabel("Batch Index")

        (bot1, top1) = ax1.get_ylim()
        (bot2, top2) = ax2.get_ylim()

        ax1.set_ylim(min(bot1,bot2), max(top1,top2))
        ax2.set_ylim(min(bot1,bot2), max(top1,top2))

        scale = 0.25
        box1 = ax1.get_position()
        box2 = ax2.get_position()
        ax1.set_position([box1.x0, box1.y0 + box1.height * scale, box1.width, box1.height * (1-scale)])
        ax2.set_position([box2.x0, box2.y0 + box2.height * scale, box2.width, box2.height * (1-scale)])

        # Put a legend below current axis
        fig.legend(*ax1.get_legend_handles_labels(), loc='upper center', bbox_to_anchor=(.5, .175), fancybox=True, ncol=2) #type: ignore

        plt.show()